/*
  Based on
  https://computing.llnl.gov/tutorials/pthreads/#ConditionVariables

  This models a "producer-consumer" problem where the consumer thread
  blocks until the producer threads fill the entire buffer. Then the
  consumer thread unblocks and "empties" the buffer.

  You can compile this program using either Linux or the pthreads-win32
  library for Windows. If you use the pthreads-win32 library for Windows,
  then you must have the files pthread.h, pthread.lib, pthreadVC2.dll in
  the same directory as this file.
*/
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>  // win32
#include <pthread.h>

#define NUM_THREADS  3  // one consumer thread and two producer threads
#define COUNT_LIMIT 12

int count = 0;
int thread_ids[3] = {0,1,2};

pthread_mutex_t count_mutex;
pthread_cond_t  count_threshold_cv;


void *inc_count(void *t)
{
  long my_id = (long)t;

  while(1)
  {
    pthread_mutex_lock(&count_mutex);
    /* Check if the "buffer" is full. */
    if (count < COUNT_LIMIT)
    {
      /* Put an item in the "buffer". */
      count++;
      /* Check again if the "buffer" is full. */
      if (count == COUNT_LIMIT)
      {
        pthread_cond_signal(&count_threshold_cv);
        printf("inc_count(): thread %ld, count = %d  Threshold reached.\n", my_id, count);
      }
    }
    else /* buffer is full, stop working */
    {
      printf("inc_count(): thread %ld, count = %d, unlocking mutex\n", my_id, count);
      pthread_mutex_unlock(&count_mutex);
      printf("inc_count(): thread %ld, count = %d, thread exiting\n", my_id, count);
      pthread_exit(NULL);
    }
    printf("inc_count(): thread %ld, count = %d, unlocking mutex\n", my_id, count);
    pthread_mutex_unlock(&count_mutex);
    /* Do some "work" so threads can alternate on mutex lock */
    //sleep(1); // posix
    Sleep(0);   // win32
  }
}


void *watch_count(void *t)
{
  long my_id = (long)t;

  printf("Starting watch_count(): thread %ld\n", my_id);
  /*
  Lock mutex and wait for signal.  Note that the pthread_cond_wait
  routine will automatically and atomically unlock mutex while it waits.
  Also, note that if COUNT_LIMIT is reached before this routine is run by
  the waiting thread, the loop will be skipped to prevent pthread_cond_wait
  from never returning.
  */
  pthread_mutex_lock(&count_mutex);
  if (count < COUNT_LIMIT)
  {
    pthread_cond_wait(&count_threshold_cv, &count_mutex);
    printf("watch_count(): thread %ld Condition signal received.\n", my_id);
    /* Do whatever is needed to empty the "buffer". */
    printf("watch_count(): thread %ld count now = %d.\n", my_id, count);
  }
  printf("watch_count(): thread %ld, count = %d, unlocking mutex\n", my_id, count);
  pthread_mutex_unlock(&count_mutex);
  printf("watch_count(): thread %ld, count = %d, thread exiting\n", my_id, count);
  pthread_exit(NULL);
  return 0;  // keep gcc happy
}


int main (int argc, char *argv[])
{
  int i;
  long t1=1, t2=2, t3=3;
  pthread_t threads[3];
  pthread_attr_t attr;

  /* Initialize mutex and condition variable objects */
  pthread_mutex_init(&count_mutex, NULL);
  pthread_cond_init (&count_threshold_cv, NULL);

  /* For portability, explicitly create threads in a joinable state */
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  pthread_create(&threads[0], &attr, watch_count, (void *)t1);
  pthread_create(&threads[1], &attr, inc_count, (void *)t2);
  pthread_create(&threads[2], &attr, inc_count, (void *)t3);

  /* Wait for all threads to complete */
  for (i=0; i<NUM_THREADS; i++)
  {
    pthread_join(threads[i], NULL);
  }
  printf ("Main(): Waited on %d  threads. Done.\n", NUM_THREADS);

  /* Clean up and exit */
  pthread_attr_destroy(&attr);
  pthread_mutex_destroy(&count_mutex);
  pthread_cond_destroy(&count_threshold_cv);
  pthread_exit(NULL);
  return 0;  // keep gcc happy
}
